#! /usr/bin/perl

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States. Recipients
# of this software assume all responsibilities associated with its operation,
# modification and maintenance. 

# This program evaluates localization runs. It can be called to evaluate 
# one or more runs. For each run, it will generate two result files:
# runname.results.table (includes summary of recall, precision, and fscore per concept and per run) and 
# runname.results.details (includes detail results for each result item submitted)
# stats file of scores across all measures and concepts

# Usage : tv15.evaluate.run.pl (runname | * | runname*)


use strict;
my(%GT, %sin_GT, %Iframes_info);
my(%run);
my($line);
my(@input);
my(%resultItems,%resultItemsTotal);
my ($results, $details);
my($runname);
my ($sin_video, $sin_shot);
my(%resultStats);

#input of iframes info
open IFRAMES_INFO, "< iacc.2.C.loc.min.max.Iframe.per.shot" or die "Can't open file $! \n";

#input GTs

open GT_SIN, "< featureF.qrels.tv15" or die "Can't open G.T of SIN $!";

open GT, "< tv15.loc.qrels" or die "Can't open G.T file $!";

#output
open STATS, "> ./results/tv15.loc.stats" or die "Can't open stats file $!";

#concepts evaluated for localization in tv15
my (@concepts) = (1003,1005,1015,1017,1019,1031,1080,1117,1261,1392);

if (scalar(@ARGV) == 0)
{
    print "usage: tv15.evaluate.run.pl (runname | * | runname*)\n";
    exit;
}

my($video, $shot, $max_iframe, $min_iframe);
##################################################################
# load the min and max index of iframes for each video and shot
while ($line = <IFRAMES_INFO>)
{
    chomp $line;
    ($video, $shot, $min_iframe, $max_iframe) = split /\s+/, $line;
    $Iframes_info{$video}{$shot}{"min"} = $min_iframe;
    $Iframes_info{$video}{$shot}{"max"} = $max_iframe;
}

#load the G.T data for SIN
while ($line = <GT_SIN>)
{
    chomp $line;
    my ($concept, undef, $file, undef, $relevance) = split /\s+/, $line;
    ($sin_video, $sin_shot) = split /_/, $file;
    $sin_video = substr($sin_video,4); 
    $sin_GT{$concept}{$sin_video}{$sin_shot} = $relevance;   
}
 print "loaded SIN G.T\n";

###################################################################################
#load the G.T data
while ($line = <GT>)
{
    chomp $line;
    my ($concept, undef, $file, undef, $relevance, $x1, $y1, $x2, $y2) = split /\s+/, $line;
    $file = substr($file,0,-5); #file_shot_iframe
    my ($GT_file, $GT_shot, $GT_iframe) = split /_/,$file;
    $GT{$concept}{$GT_file}{$GT_iframe}{$relevance} = "$x1 $y1 $x2 $y2";          
}
 print "loaded G.T\n";
###################################################################################

foreach $runname(@ARGV)
{
    @input = split /\//, $runname;
   
    $results = "./results/$input[-1]".".results.summary";
    $details = "./results/$input[-1]".".results.details";

    #results table file per run (showing results per concept and per run)
    open RESULTS,">$results" or die "Can't open results file $!";

    #diagnostic details file for each result item
    open DETAILS,">$details" or die "Can't open details file $!";

    #load run
    open RUN , "< $runname" or die "can't open run file $!";

    #iniitalize the run hash for the new run
    %run = ();

   while ($line = <RUN>)
   {
       chomp $line;
       my ($sub_con, $sub_file, $sub_iframe, $sub_x1, $sub_y1, $sub_x2, $sub_y2) = split /\s+/, $line;
       $sub_con = $sub_con < 1000 ? $sub_con + 1000 : $sub_con;
       $sub_iframe = $sub_iframe + 0;       
       $run{$sub_con}{$sub_file}{$sub_iframe}{"$sub_x1 $sub_y1 $sub_x2 $sub_y2"} = "$sub_x1 $sub_y1 $sub_x2 $sub_y2";
   }
   print "loaded Run $runname\n";
####################################################################################
# main loop on concepts to calc. pixel avg R and P and frame R & P. 
# saving the results in results.table and details files
    &print_details_header();

    my ($con, $avg_conR, $avg_conP, $avg_conF, $avg_Box, $frameR, $frameP, $frameF);
    my($Total_conR) = 0;
    my($Total_conP) = 0;
    my($Total_conF) = 0;
    my($Total_conBox) = 0;
    my($Total_frameR) = 0;
    my($Total_frameP) = 0;
    my($Total_frameF) = 0;

    my($concepts_count) = scalar(@concepts);
    foreach $con(@concepts)
    {
	($avg_conR, $avg_conP, $avg_conF, $avg_Box) = calc_avg_pixel_PR_givenConcept($con); #Avg pixel Recall ,Precision , & Fscore / concept, write results.details file
	($frameR, $frameP, $frameF) = calc_temporal_retrieval_givenConcept($con); #Frame Recall, Precision, & Fscore / concept

	#Any new measurements to be reported can be added here to appear in the result table summary file
	$Total_conR += $avg_conR;
	$Total_conP += $avg_conP;
	$Total_conF += $avg_conF;
	$Total_conBox += $avg_Box;
	$Total_frameR += $frameR;
	$Total_frameP += $frameP;
	$Total_frameF += $frameF;
	$resultItems{"mean_pixel_recall"}{$con} = $avg_conR;
	$resultItems{"mean_pixel_precision"}{$con} = $avg_conP;
	$resultItems{"mean_pixel_fscore"}{$con} = $avg_conF;       
	$resultItems{"iframe_recall"}{$con} = $frameR;
	$resultItems{"iframe_precision"}{$con} = $frameP;
	$resultItems{"iframe_fscore"}{$con} = $frameF;
	$resultItems{"false_positives"}{$con} = get_submitted_FPs_givenConcept($con);
	$resultItems{"Total_relevant_Iframes"}{$con} =  get_GT_totalPositive_givenConcept($con);
	$resultItems{"Total_Submitted_Iframes"}{$con} =  get_submitted_totalIframes_givenConcept($con);
	$resultItems{"TP_Submitted_Iframes"}{$con} =  get_run_GT_overlapping_givenConcept($con);
	$resultItems{"mean_box_per_Iframe"}{$con} = $avg_Box;

	#store for stats
	push @{$resultStats{"mean_pixel_recall"}{$con}} , $avg_conR;
	push @{$resultStats{"mean_pixel_precision"}{$con}} , $avg_conP;
	push @{$resultStats{"mean_pixel_fscore"}{$con}} , $avg_conF;
	push @{$resultStats{"iframe_recall"}{$con}} , $frameR;
	push @{$resultStats{"iframe_precision"}{$con}} , $frameP;
	push @{$resultStats{"iframe_fscore"}{$con}} , $frameF;
    }

    $resultItemsTotal{"mean_pixel_recall"} = $Total_conR/$concepts_count;
    $resultItemsTotal{"mean_pixel_precision"} = $Total_conP/$concepts_count;
    $resultItemsTotal{"mean_pixel_fscore"} = $Total_conF/$concepts_count;   
    $resultItemsTotal{"iframe_recall"} = $Total_frameR/$concepts_count;
    $resultItemsTotal{"iframe_precision"} = $Total_frameP/$concepts_count;
    $resultItemsTotal{"iframe_fscore"} = $Total_frameF/$concepts_count;
    $resultItemsTotal{"Total_relevant_Iframes"} =  "---";
    $resultItemsTotal{"Total_Submitted_Iframes"} =  "---";
    $resultItemsTotal{"TP_Submitted_Iframes"} = "---";
    $resultItemsTotal{"false_positives"} = "---";
    $resultItemsTotal{"mean_box_per_Iframe"} = $Total_conBox/$concepts_count;
    
    #write results summary in results.table file
    write_resultsTable($input[-1]);

    close RESULTS;
    close DETAILS;
}

write_stats();

####################################################################################
############# returns the temporal recall and precision given a concept ####################
sub calc_temporal_retrieval_givenConcept()
{
    my ($concept) = @_;   
    my ($recall) = (get_GT_totalPositive_givenConcept($concept) == 0) ? 0 : get_run_GT_overlapping_givenConcept($concept) / get_GT_totalPositive_givenConcept($concept);
    my ($precision) = (get_submitted_totalIframes_givenConcept($concept) == 0) ? 0 : get_run_GT_overlapping_givenConcept($concept) / get_submitted_totalIframes_givenConcept($concept);
    my ($fscore) = calc_fscore($recall,$precision);
    
    return($recall, $precision, $fscore);   
}
############################################################################################

##################### returns the average Recall and precision of pixel localization across all files for a given concept ###################
sub calc_avg_pixel_PR_givenConcept()
{    
    my ($concept) = @_;
    my($file);
    my($iframe);
    my($c, $boxes_count, $c_FP) = 0;
    my($R,$P,$F,$id);
    my($Rtotal) = 0;
    my($Ptotal) = 0;
    my($Ftotal) = 0;
    my($boxes_avg) = 0;
    my($max_F) = 0;
    my($final_R) = 0;
    my($final_P) = 0;
    my($mean_box,$flag);
    my($sx1,$sy1,$sx2,$sy2,$gx1,$gx2,$gy1,$gy2, $t,$max_f_index, $line, @tmp_str );
    
   
    my $debug;
    foreach $file(sort keys %{$run{$concept}})
    {
	foreach $iframe(sort keys %{$run{$concept}{$file}})
	{
	   	   	   
		$max_F = -1;
		$boxes_count = 0;
		$t = 0;
		$max_f_index = $t;
		@tmp_str = ();
		$c_FP = 0;
		foreach $id(sort keys %{$run{$concept}{$file}{$iframe}})
		{ 
		    $boxes_count++;
		     
                     # NOT judged in localization / SIN		    
		     if (get_relevance($concept, $file, $iframe) == -1) 
		     {
			 if (get_sin_rel($concept,$file,$iframe) == -2)
			 {#invalid I-frame (out of boundary)
			     my($sx1,$sy1,$sx2,$sy2) = split/\s+/, $run{$concept}{$file}{$iframe}{$id}; #get submitted locations
			     printf DETAILS "%4d %8d %5d %4s %4s %4s %4s %4d %4d %4d %4d %9.4f %8.4f %8.4f %15s\n",$concept, $file, $iframe, "*","*","*","*", $sx1, $sy1, $sx2, $sy2,0,0,0,"Invalid-Iframe";
			 }
			 else
			 {
			     my($sx1,$sy1,$sx2,$sy2) = split/\s+/, $run{$concept}{$file}{$iframe}{$id}; #get submitted locations
			     printf DETAILS "%4d %8d %5d %4s %4s %4s %4s %4d %4d %4d %4d %9.4f %8.4f %8.4f %15s\n",$concept, $file, $iframe, "na","na","na","na", $sx1, $sy1, $sx2, $sy2,0,0,0,"Not-Judged";			 
			 }
			 #next;
		     }
		    
		     else #Judged
		     {
			 ($R, $P, $F,$flag) = calc_pixel_retrieval($concept, $file, $iframe, $id);
	       
			 if ($flag == 0) #No target in iframe OR non-relevant SIN video
			 {			   			    
			    $Rtotal += $R;
			    $Ptotal += $P;
			    $Ftotal += $F;			    
			    $c_FP++;
			    
			    ($sx1,$sy1,$sx2,$sy2) = split/\s+/, $run{$concept}{$file}{$iframe}{$id};
			    printf DETAILS "%4d %8d %5d %4d %4d %4d %4d %4d %4d %4d %4d %9.4f %8.4f %8.4f %15s\n",$concept, $file, $iframe, 0,0,0,0, $sx1, $sy1, $sx2, $sy2,0,0,0,"No-Target";
			    
			 }
			
			 # Keep the R,P,& F scores of the box with max f-score value in case of multiple boxes per iframe
			 elsif ($F > $max_F)
			 {
			     $max_F = $F;
			     $final_R = $R;
			     $final_P = $P;
			     ($sx1,$sy1,$sx2,$sy2) = split/\s+/, $run{$concept}{$file}{$iframe}{$id};
			     ($gx1, $gy1, $gx2, $gy2) = get_location($concept, $file, $iframe);
			    
			     $tmp_str[$t] = sprintf "%4d %8d %5d %4d %4d %4d %4d %4d %4d %4d %4d %9.4f %8.4f %8.4f",$concept, $file, $iframe, $gx1, $gy1, $gx2, $gy2, $sx1, $sy1, $sx2, $sy2, $R, $P, $F;			     
			     $max_f_index = $t;
			     $t++;
			 }    
			 else
			 {
			     ($sx1,$sy1,$sx2,$sy2) = split/\s+/, $run{$concept}{$file}{$iframe}{$id};
			     ($gx1, $gy1, $gx2, $gy2) = get_location($concept, $file, $iframe);
			     $tmp_str[$t] = sprintf "%4d %8d %5d %4d %4d %4d %4d %4d %4d %4d %4d %9.4f %8.4f %8.4f",$concept, $file, $iframe, $gx1, $gy1, $gx2, $gy2, $sx1, $sy1, $sx2, $sy2, $R, $P, $F;			      
			     $t++;
			 }
			
		     } #judged
	        } #boxes per iframe

		my $i = 0;
		foreach $line(@tmp_str)
		{
		    if ($i == $max_f_index)
		    {
			printf DETAILS "$line 1\n";
		    }
		    else
		    {
			printf DETAILS "$line   Not-selected\n";
		    }
		    $i++;
		}
	
		if ($max_F > -1)
		{
		    $boxes_avg += $boxes_count;
		    $Rtotal += $final_R;
		    $Ptotal += $final_P;
		    $Ftotal += $max_F;
		    $c++; # update i-frames counter
  	    	    
		}		
		elsif ($max_F == -1 && $c_FP > 0) #count once any FP (even if iframe was submitted more than once)
		{
		    $c++;
		}
	    
	} #iframe
    } #video file
    
   # print "debug counter = $debug, c counter = $c \n";
    if ($c == 0 )
    {
	 return(0,0,0,0);
    }
    else
    {
#	print "for concept $concept ---> counter of boxes = $c\n";
	
	return($Rtotal/$c , $Ptotal/$c, $Ftotal/$c, $boxes_avg/$c);
    }
    
}

###############################################################################################################################################
#calc a submitted bounding box recall and precision (overlapping area/G.T area and overlapping area/submitted area)
sub calc_pixel_retrieval()
{ 
    my($concept, $file, $iframe, $id) = @_;
    my($recall, $precision, $fscore);
    my($sx1,$sy1,$sx2,$sy2) = split/\s+/, $run{$concept}{$file}{$iframe}{$id}; #get submitted locations

    if (get_relevance($concept, $file, $iframe) == 1) #Target Exists
    {
	my($gx1, $gy1, $gx2, $gy2) = get_location($concept, $file, $iframe); #get G.T locations
	if ( ($gx1 - $gx2) == 0 || ($gy1 - $gy2) == 0 )
	{
	   print "ERROR!! Found a G.T relevant with a non-box coordinates at $concept and file $file frame $iframe\n";
	   exit;
	}
	$recall = (max(0,min($gx2,$sx2) - max($gx1,$sx1) + 1) * max(0,min($gy2,$sy2)-max($gy1,$sy1) + 1) ) / ((($gx2-$gx1)+1) * (($gy2-$gy1)+1));

	$precision = (max(0,min($gx2,$sx2) - max($gx1,$sx1) + 1) * max(0,min($gy2,$sy2)-max($gy1,$sy1) + 1) ) / ((($sx2-$sx1)+1) * (($sy2-$sy1)+1));

	$fscore = calc_fscore($recall, $precision);
	
	return($recall, $precision,$fscore,1);
    }
    elsif (get_relevance($concept, $file, $iframe) == 0) #No target exists
    {
    	return(0,0,0,0);
    }
 
}

#################################################################################################################################################

# returns the x1,y1,x2,y2 location in the ground truth (if 4 values are 0, then there is no target (relevance is 0)
sub get_location()
{
    my ($concept, $file, $iframe) = @_;   
    my ($rel);
    my($loc);
    foreach $rel(sort keys %{$GT{$concept}{$file}{$iframe}})
    {
	$loc = $GT{$concept}{$file}{$iframe}{$rel};
	my ($x1, $y1, $x2, $y2) = split /\s+/, $loc;
	return($x1, $y1, $x2, $y2);
    }
}

# returns 1 or 0 for this concept,file and iframe from the G.T
sub get_relevance()
{
    my ($concept, $file, $iframe) = @_;   
    my($rel);
    if (isJudged($concept, $file, $iframe) == 1) #judged
    {
	foreach $rel(sort keys %{$GT{$concept}{$file}{$iframe}})
	{
	    return($rel);  #return 1 or 0		
	}
	return(0); #FP SIN video
    }
    elsif (isJudged($concept, $file, $iframe) == 0)
    {
	return(-1); #not judged in Loc or SIN 
    }

}

#checks if an iframe was localized or not (return 1 if judged, otehrwise 0)
sub isJudged()
{
    my ($concept, $file, $iframe) = @_;   
    if (exists ($GT{$concept}{$file}{$iframe}{"1"}))
    {
    	    return(1); #Judged 1 in Localization
    }
    
    if (exists ($GT{$concept}{$file}{$iframe}{"0"})) 
    {
    	    return(1); #judged 0 in Localization
    }
         
    if (get_sin_rel($concept,$file,$iframe) == 0) 
    {
    	    return(1); #Judged 0 in SIN
    }

    if (get_sin_rel($concept,$file,$iframe) == -1) 
    {
    	    return(0); #not judged in SIN
    }
    
     if (get_sin_rel($concept,$file,$iframe) == -2) 
    {
    	    return(0); #not judged in SIN (frame out of boundary)
    }
    
    return(0); #not judged in localization    
}

# returns the number of TP(target) iframes in G.T across all files for a given concept
sub get_GT_totalPositive_givenConcept()
{
    my ($concept) = @_;
    my($file);
    my($iframe);
    my($rel);
    my($c) = 0;
   # print "$concept\n";
    foreach $file(sort keys %{$GT{$concept}})
    {
	foreach $iframe(sort keys %{$GT{$concept}{$file}})
	{
	   foreach $rel(sort keys %{$GT{$concept}{$file}{$iframe}})
	   { 
	       if ($rel == 1)
	       {
		   $c++;
	       }
	   }
	}
    }
    return($c);
}

# returns the number of FP iframes in submissions across all files for a given concept. In case of duplicate iframes for a given file, we count it only 1 time
sub get_submitted_FPs_givenConcept()
{    
    my ($concept) = @_;
    my($file);
    my($iframe);
    my($c) = 0;
   # print "$concept\n";
    foreach $file(sort keys %{$run{$concept}})
    {
	foreach $iframe(sort keys %{$run{$concept}{$file}})
	{
	    #if iframe is judged 0 (non-relevant
	    if (&get_relevance($concept, $file, $iframe) == 0)
	    {
		$c++;	   
	    }
	   
	}
    }
    return($c);
    
}


# returns the number of submitted iframes (in a run) across all files for a given concept. In case of duplicate iframes for a given file, we count it only 1 time
sub get_submitted_totalIframes_givenConcept()
{    
    my ($concept) = @_;
    my($file);
    my($iframe);
    my($c) = 0;
   # print "$concept\n";
    foreach $file(sort keys %{$run{$concept}})
    {
	foreach $iframe(sort keys %{$run{$concept}{$file}})
	{
	    #if iframe is judged in Localization
	    if (&isJudged($concept, $file, $iframe) == 1)
	    {
		$c++;	   
	    }
	   
	}
    }
    return($c);
    
}




# returns the overlapping between TP (target) iframes in G.T AND submitted run for a given concept
sub get_run_GT_overlapping_givenConcept()
{
    my ($concept) = @_;
    my($file);
    my($iframe);
    my($c) = 0;
    foreach $file(sort keys %{$run{$concept}})
    {
	foreach $iframe(sort keys %{$run{$concept}{$file}})
	{	 
		if (exists $GT{$concept}{$file}{$iframe}{"1"})
		{
		    $c++;
	
		}	
	}
    }
    return($c)
}


sub calc_fscore()
{
    my($recall, $precision) = @_;
    my($fscore);
    if (($recall + $precision) == 0)
    {
	$fscore = 0;
    }
    else
    {
	$fscore = (2 * $precision * $recall) / ($precision + $recall);
    }
   
    return($fscore);
}

sub max()
{
    my ($x, $y) = @_;
    $x >= $y ? return($x) : return($y);
}

sub min()
{
    my ($x, $y) = @_;
    $x <= $y ? return($x) : return($y);
}

sub print_details_header()
{
    printf DETAILS "%10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %15s\n\n","Concept FileID Iframe  Gx1  Gy1  Gx2  Gy2  Sx1  Sy1  Sx2  Sy2  Recall   Precsion   Fscore   Description";
    
}

sub median {
    my @array = sort {$a <=> $b} @_;
    my $median;

    if ($#array % 2 == 0) 
    { #size of aray -1
        $median = $array[($#array / 2)];
    }
    else 
    {
        $median = $array[int($#array / 2)] + (($array[int($#array / 2) + 1] - $array[int($#array / 2)]) / 2);
    }
    return($median);
}

sub write_resultsTable()
{
    my($run) = @_;
    my($i,$con);
    my(@record);

    ######################################
    print RESULTS "**** Note: duplicate submitted I-frames are counted in this summary table once (for both TPs and FPs). ****\n\n";

    ####### print concept IDs ############
    printf RESULTS "%66s";
    foreach $con(sort keys %{$resultItems{"mean_pixel_recall"}})
    {
	printf RESULTS "%8d",$con;
    }
    printf RESULTS "%15s","Mean_Per_Run";
    print RESULTS "\n\n";
    
    ############ print result items per concept
    foreach $i(sort keys %resultItems)
    {
	printf RESULTS "%40s %25s",$run, $i;
	foreach $con(sort keys %{$resultItems{$i}})
	{
	    if (($i eq "Total_relevant_Iframes") || ($i eq "Total_Submitted_Iframes") || ($i eq "TP_Submitted_Iframes") || ($i eq "false_positives"))
	    {
		printf RESULTS "%8d",$resultItems{$i}{$con};
	    }
	    else
	    {
		printf RESULTS "%8.4f",$resultItems{$i}{$con};
	    }
	}
	if (($i eq "Total_relevant_Iframes") || ($i eq "Total_Submitted_Iframes") || ($i eq "TP_Submitted_Iframes") || ($i eq "false_positives"))
	{
	     printf RESULTS "%8s",$resultItemsTotal{$i};
	}
	else
	{
	    printf RESULTS "%8.4f",$resultItemsTotal{$i};
	}
	print RESULTS "\n";
    }

}

#write stats file (measure concept min max median score1 score2 ...scoreN)
sub write_stats()
{
    my($i,$con,$res);
    my(@sortedRes);
    foreach $i(sort keys %resultStats)
    {
	foreach $con(sort keys %{$resultStats{$i}})
	{
	    @sortedRes = sort @{$resultStats{$i}{$con}};	    
	                                                          #min            #max            #median
	    printf STATS "%25s %8d %8.4f %8.4f %8.4f", $i, $con, $sortedRes[0], $sortedRes[-1], median(@sortedRes) ;
	    foreach $res(sort @{$resultStats{$i}{$con}})
	    {
		printf STATS "%8.4f",$res;
	    }
	    print STATS "\n";
	}
    }
    close STATS;
}


sub get_sin_rel()
{    
    my ($concept, $video, $iframe) = @_;
    my ($shot);

    if ($iframe == 0)
    {
	print "Error -> $concept $video $iframe\n";
	exit;
    }
    $shot = lookup_sin_shot($video,$iframe);   
    
    if ($shot == -1)
    {
	return(-2); #error code for out of bound iframe!!
	#exit;
    }
    #return relevancy of this shot == relevancy of the iframe in question
    return($sin_GT{$concept}{$video}{$shot}); 
    
}

sub lookup_sin_shot()
{
     my($video, $iframe) = @_;
     my($vid,$shot);
     
     
     foreach $shot (sort keys %{$Iframes_info{$video}})
     { 
	 if ($Iframes_info{$video}{$shot}{"min"} <= $iframe &&  $Iframes_info{$video}{$shot}{"max"} >= $iframe)
	 {
		 return($shot);
	 }
     }
     
     return(-1); #error! video or iframe is out of boundary for the tested collection
}
